module(..., package.seeall)
require "defines"

prices =
{
  ["coal"]=5,
  ["basic-transport-belt"]=5,
  ["basic-transport-belt-to-ground"]=20,
  ["fast-transport-belt"]=50,
  ["fast-transport-belt-to-ground"]=200,
  ["basic-splitter"]=25,
  ["fast-splitter"]=50,
  ["burner-inserter"]=10,
  ["basic-inserter"]=10,
  ["long-handed-inserter"]=15,
  ["fast-inserter"]=20,
  ["smart-inserter"]=35,
  ["red-wire"]=2,
  ["green-wire"]=2,
  ["wooden-chest"]=5,
  ["iron-chest"]=10,
  ["smart-chest"]=20,
  ["stone-furnace"]=10,
  ["steel-furnace"]=50,
  ["electric-furnace"]=70,
  ["offshore-pump"]=10,
  ["pipe"]=5,
  ["pipe-to-ground"]=20,
  ["boiler"]=15,
  ["steam-engine"]=50,
  ["small-electric-pole"]=10,
  ["medium-electric-pole"]=50,
  ["big-electric-pole"]=100,
  ["substation"]=150,
  ["assembling-machine-1"]=30,
  ["assembling-machine-2"]=50,
  ["assembling-machine-3"]=100,
  ["basic-mining-drill"]=50,
  ["burner-mining-drill"]=10,
  ["pumpjack"]=50,
  ["oil-refinery"]=100,
  ["chemical-plant"]=50,
  ["storage-tank"]=50,
  ["storage-tank"]=40,
  ["small-pump"]=20,
  ["logistic-robot"]=150,
  ["logistic-chest-passive-provider"]=50,
  ["logistic-chest-requester"]=50,
}

function makeoffer(item)
  assert(prices[item])
  return { price={{"coin", prices[item]}}, offer={type="give-item", item=item}}
end

gt = game.gettext
game.disablerecipegroups()
game.disableminimap()
game.mapsettings.pollution.enabled = false

-- The utility for managing tight-spot campaign levels
TightSpot =
{
}

function TightSpot:new(level)
  obj = {}
  setmetatable(obj, self)
  self.__index = self
  obj.level = level
  -- Part of the level data that needs to be serialised to preserve game state
  obj.context = {}
  -- table of demand indexed by demand key. The original demand table can't be
  -- indexed, as it would cause non-deterministic order of demands when iterated
  obj.demand = {}
  for index, item in pairs(level.demand) do
    obj.demand[item.item] = item.price
  end
  return obj
end

function TightSpot:getrequiredbalance()
  if game.difficulty == defines.difficulty.easy then
    return self.level.requiredbalance.easy
  elseif game.difficulty == defines.difficulty.normal then
    return self.level.requiredbalance.normal
  else
    return self.level.requiredbalance.hard
  end
end

function TightSpot:initlevel()
  local character = game.player.character
  game.player.setcontroller{type=defines.controllers.god}
  if character then character.destroy() end
  game.player.force.manualminingspeedmodifier = 1000
  local level = self.level
  self.context.debt = level.money
  self.context.income = 0
  for x=level.area[1][1],level.area[2][1],1 do
    for y=level.area[1][2],level.area[2][2],1 do
      local entities = game.findentities{{x + 0.1, y + 0.1}, {x + 0.9, y + 0.9}}
      local shouldcover = true;
      for index, item in pairs(entities) do
        if item.type == "resource" or
           item.type == "tree" then
          item.minable = false
          shouldcover = false;
        end
      end
      if shouldcover and game.canplaceentity{name = "wall", position = {x + 0.5, y + 0.5}} then
        game.createentity{name = "wall", position = {x, y}}
      end
    end
  end
  local market = game.getentitybytag("market")
  for index, item in pairs(level.offers) do
    market.addmarketitem(item)
  end
  game.player.insert{name= "steel-axe", count=1}
  game.player.insert{name = "coin", count = level.money}
  game.player.force.disableallprototypes()
  local recipeList = game.player.force.recipes
  for index, item in pairs(level.recipes) do
    recipeList[item].enabled = true
  end
  self:initspendingframe()
  self:updatespending()
end

function TightSpot:initspendingframe()
  local frame = game.player.gui.top.add{type="frame", name="tightspot_frame", caption = "Level " .. self.level.levelnumber, direction="vertical"}
  local table = frame.add{type="table", name="table", colspan=2}
  local available = game.player.getitemcount("coin")
  local debt = self.context.debt
  table.add{type="label", caption=gt("money") .. ":", style="caption_label_style"}
  table.add{type="label", caption=""}
  table.add{type="label", caption=gt("balance") .. ":"}
  table.add{type="label", caption=debt - available, name="balance"}
  table.add{type="label", caption=gt("required") .. ":"}
  table.add{type="label", caption=self:getrequiredbalance(), name="required"}

  local level = self.level
  --local timeflow = frame.add{type="flow", name="timeflow"}
  table.add{type="label", caption=gt("time") .. ":"}
  table.add{type="label", caption=util.formattime(level.time)}

  frame.add{type="label", caption=gt("demand") .. ":", style="caption_label_style"}
  local demandtable = game.player.gui.top.tightspot_frame.add{type="table", name="demandtable", colspan=2}
  
  for index, item in pairs(level.demand) do
    demandtable.add{type="label", caption=game.getlocaliseditemname(item.item)}
    demandtable.add{type="label", caption=item.price}
  end

  game.player.gui.top.add{type = "button", name="tightspot_start", caption=gt("start")}
end

function TightSpot:initprogressframe()
  local progressframe = game.player.gui.top.add{type = "frame", name="tightspot_progress_frame", caption=gt("progress"), direction="vertical"}
  progressframe.direction = "vertical"
  progressframe.add{type="progressbar", size="10", name="tightspot_progressbar"}
  local table = progressframe.add{type="table", colspan="2", name="table"}
  table.add{type="label", caption=gt("time") .. ":"}
  table.add{type="label", name="time"}
  table.add{type="label", caption=gt("speed") .. ":"}
  table.add{type="label", name="speed"}
  local controlflow = progressframe.add{type="flow"}
  controlflow.add{type="button", caption="+", name="faster"}
  controlflow.add{type="button", caption="-", name="slower"}
  self:updatespeedlabel()
end

function TightSpot:updateprogress()
  if (game.tick - self.context.simulationstartedat) % 60 ~= 0 then
    return
  end
  local tightspot_progress_frame = game.player.gui.top.tightspot_progress_frame
  local time = game.tick - self.context.simulationstartedat
  tightspot_progress_frame.tightspot_progressbar.value = time / level.time
  tightspot_progress_frame.table.time.caption = util.formattime(time)
end

function TightSpot:updatespending()
  local table = game.player.gui.top.tightspot_frame.table
  local available
  if game.player.controllertype == defines.controllers.god then
    available = game.player.getitemcount("coin")
   else
    available = self.context.available
  end
  local debt = self.context.debt
  table.balance.caption = available - debt + self.context.income
end

function TightSpot:getmissingtowin()
  local available = self.context.available
  local debt = self.context.debt
  local required = self:getrequiredbalance()
  return required + debt - self.context.income - available
end

function TightSpot:sellback()
  local level = self.level
  local player = game.player
  local addition = 0
  coinsleft = player.getitemcount("coin")
  player.removeitem{name = "coin", count = coinsleft}
  for index, item in pairs(level.offers) do
    local count = player.getitemcount(item.offer.item)
    addition = addition + count * item.price[1][2]
  end
  game.createentity{name = "flying-text", position = player.position, text = "+" .. addition, color = {g = 1}}
  self.context.debt = self.context.debt - addition - coinsleft
  player.clearitemsinside()
end

function TightSpot:startlevel()
  game.save("before-run")
  game.player.gui.top.tightspot_start.destroy()
  self:sellback()
  game.speed = 32
  self:initprogressframe()
  game.player.gui.top.add{type = "button", name="tightspot_reload", caption=gt("reload")}
  self.context.available = game.player.getitemcount("coin")
  game.player.setcontroller{type=defines.controllers.ghost}
  self.context.simulationstartedat = game.tick
  local level = self.level
  local entities = game.findentities(level.area)
  for index, item in pairs(entities) do
    item.active = true
  end
  self:updatespending()
end

function TightSpot:showleveldescription()
  local level = self.level
  if level.description ~= nil then
    game.showmessagedialog(level.description)
  end
end

function TightSpot:updatespeedlabel()
  game.player.gui.top.tightspot_progress_frame.table.speed.caption = game.speed
end

function TightSpot:checklevel(event)
  if event.name == defines.events.onguiclick then
    if event.element.name == "slower" and game.speed > 1 then
      game.speed = game.speed / 2
      self:updatespeedlabel()
    elseif event.element.name == "faster" and game.speed < 64 then
      game.speed = game.speed * 2
      self:updatespeedlabel()
    elseif event.element.name == "tightspot_reload" then
      game.load("before-run")
      return false
    end
    return false
   end

  local resultchest = game.getentitybytag("result-chest")
  local contents = resultchest.getinventory(defines.inventory.chest).getcontents()
  local addition = 0
  for itemname, count in pairs(contents) do
    local price = self.demand[itemname]
    if price ~= nil then
      addition = addition + price * count
    end
  end
  if addition ~= 0 then
    game.createentity{name = "flying-text", position = resultchest.position, text = "+" .. addition, color = {g = 1}}
    self.context.income = self.context.income + addition
    self:updatespending()
  end
  self:updateprogress()
  resultchest.clearitemsinside()
  return (game.tick - self.context.simulationstartedat) >= level.time
end

function TightSpot:import(context)
  self.context = context
end

function TightSpot:export()
  return self.context
end

function landprice(level, position)
  return math.abs(position.x - level.center.x) + math.abs(position.y - level.center.y)
end

storytable =
{
  {
    {
      action = function()
        levelengine:initlevel()
      end
    },
    {
      action = function()
        if levelengine.level.showrules ~= nil then
          game.showmessagedialog(gt("welcome"))
          game.showmessagedialog(gt("rules1"))
          game.showmessagedialog(gt("rules2"))
          game.showmessagedialog(gt("rules3"))
        end
      end
    },
    {
      action = function()
        levelengine:showleveldescription()
      end
    },
    {
      update =
      function(event)
        levelengine:updatespending()
        if event.name == defines.events.onbuiltentity then
          if event.createdentity.name == "wall" then
            if #game.findentitiesfiltered{area = {{event.createdentity.position.x - 0.4,
                                                   event.createdentity.position.y - 0.4},
                                                  {event.createdentity.position.x + 0.4,
                                                   event.createdentity.position.y + 0.4}},
                                          type = "resource"} ~= 0 then
              event.createdentity.destroy()
              game.player.insert{name = "wall", count = 1}
              game.player.print("This area is not resellable")
              return
            end
            local price = landprice(levelengine.level, event.createdentity.position)
            game.createentity{name = "flying-text", position = event.createdentity.position, text = "+" .. price, color = {g = 1}}
            game.player.insert{name = "coin", count = price}
          else
            event.createdentity.active = false
          end
        end
        if event.name == defines.events.onpreplayermineditem and
           event.entity.name == "wall" then
          local position = event.entity.position
          local price = landprice(levelengine.level, event.entity.position)
          game.createentity{name = "flying-text", position = event.entity.position, text = "-" .. price, color = {r = 1}}
          local available = game.player.getitemcount("coin")
          game.player.removeitem{name = "coin", count = price}
          if available < price then
            levelengine.context.debt = levelengine.context.debt + 5000
            game.player.insert{name = "coin", count = 5000 - price + available}
            game.player.print("Borrowed 5000 coins to pay for the land")
          end
        end
      end,
      condition =
      function(event)
        return event.name == defines.events.onguiclick and event.element.name == "tightspot_start"
      end,
      action =
      function()
        levelengine:startlevel()
      end
    },
    {
      condition =
      function(event)
        return levelengine:checklevel(event)
      end,
      action =
      function()
        local missing = levelengine:getmissingtowin()
        if missing > 0 then
          game.showmessagedialog(gt("lost-message", missing))
        else
          game.showmessagedialog(gt("won-message", -missing))
        end
      end
    },
    {
      condition =
      function()
        if levelengine:getmissingtowin() > 0 then
          game.load("before-run")
          return false
        end
        return true
      end,
      action = function() end
    }
  }
}
